# pyobjects.py
# -*- coding: utf8 -*-
# Copyright 2008 Simon Cross <hodgestar@gmail.com>
# GPL - see COPYING for details

"""
"Everything You Didn't Want to Known About Python Objects"
 - in which I chronicle my recent findings in the
   depths of Object/object.c.

Talk given to the Cape Town Python Users Group on 21 June 2008.
"""

from slides import Lecture, Slide, Bullet, PRE, Image, SubBullet, CODE, PYCODE, PYCON, cgi

PYTHON_TRUNK_CHECKOUT = "/your/python/trunk"

#
# Customisation of Python slides for this talk
#

class MyLecture(Lecture):
    """Allow list of CSS files for pygments rendering.
       """
    def getHeader(self):
        """Return the header for HTML pages."""
        result = '<html>\n<head>\n<title>%s</title>\n' % cgi.escape(self.name)
        if self.css:
            for css_file in self.css:
                result += '<link rel="stylesheet" type="text/css" href="%s" />\n' % css_file
        return result + '</head>\n<body>\n'

class CenteredSlide(Slide):
    """Centered slides."""
    def toHTML(self):
        html = super(CenteredSlide, self).toHTML()
        return "<div style='margin: auto; width: 50%;'>" + html + "</div>"

Slide = CenteredSlide

class TitleSlide(Slide):
    """Centered title slides."""
    def toHTML(self):
        title = '<h1>%s</h1>' % cgi.escape(self.title)
        bullets = []
        for bullet in self.bullets:
            bullets.append(bullet.toHTML())
        return "<div style='margin: auto; width: 50%;'>" + title + self.start + "\n".join(bullets) + self.end + "</div>"

class BulletlessSlide(Slide):
    """Slide without bullets."""
    start = "<ul style='list-style-type: none;'>\n"
    end = "</ul>\n"

class BulletlessTitleSlide(TitleSlide):
    """Title slide without bullets."""
    start = "<ul style='list-style-type: none;'>\n"
    end = "</ul>\n"

class URL(object):
    """A URL with both description and link."""
    def __init__(self, desc, url):
        self.url = url
        self.desc = desc

    def toHTML(self):
        return '<a href="%s">%s</a>' % (self.url, self.desc)

#
# PyObjects Talk
#

l = MyLecture(
    "Everything You Didn't Want to Know about Python Objects",
    BulletlessTitleSlide("Everything You Didn't Want to Know about Python Objects",
        Bullet("in which I chronicle my recent findings in the depths of Object/object.c"),
    ),
    Slide("Out With The Old ...",
        Bullet("Old style:", PYCON("""
            >>> class Old:
            ...   pass
            ... 
            >>> type(Old), type(Old())
            (<type 'classobj'>, <type 'instance'>)
            >>> Old.__class__
            Traceback (most recent call last):
              File "<stdin>", line 1, in <module>
            AttributeError: class Old has no attribute '__class__'
            >>> Old().__class__
            <class util.Old at 0xb7d4e62c>
            """)),
        Bullet("New Style:", PYCON("""
            >>> class New(object):
            ...   pass
            ... 
            >>> type(New), type(New())
            (<type 'type'>, <class '__main__.New'>)
            >>> New.__class__
            <type 'type'>
            >>> New().__class__
            <class 'util.New'>
            """)),
        Bullet("Old + __metaclass__ == New"),
    ),
    Slide("Gambolling With Slots",
        Bullet("Adding slots:", PYCON("""
            >>> class FooBar(object):
            ...   def __str__(self):
            ...     return "foobar"
            ...
            >>>   def __unicode__(self):
            ...     return unicode("foobar")
            ...
            >>> str(FuBar), str(FuBar())
            ("<class 'util.FuBar'>", 'fubar')
            >>> unicode(FuBar)
            Traceback (most recent call last):
              File "<stdin>", line 1, in <module>
            TypeError: unbound method __unicode__() must be called with FuBar instance as first argument (got nothing instead)
            >>> unicode(Uni())
            u'fubar'
        """)),
        Bullet("Same thing happens with old-style classes.")
    ),
    BulletlessSlide("object.c",
        Bullet(Image("tabs.png")),
    ),
    Slide("Strings",
        Bullet("Examples:", PYCON("""
            >>> x = chr(13)
            >>> y = chr(13)
            >>> id(x), id(y)
            (3082874688L, 3082874688L)
            >>> x = chr(13) + chr(14)
            >>> y = chr(13) + chr(14)
            >>> id(x), id(y)
            (3077159616L, 3077159584L)
            >>> x = "a"*20
            >>> y = "a"*20
            >>> z = "a"*21
            >>> w = "a"*21
            >>> id(x), id(y), id(z), id(w)
            (3079663488L, 3079663488L, 3079721328L, 3079721568L)
        """)),
        Bullet("Single characters and empty string held in permament cache (Objects/stringobject.c)."),
        Bullet("Constants strings of size <= 20 interned by parser."),
        Bullet("PyString_InternInPlace (PyUnicode_InternInPlace in 3.0) results in strings being stored in internal dictionary.",
            SubBullet(
                Bullet("Reference in interned not counted in reference count (nasty hack to hide this from users)."),
            ),
        ),
        Bullet("Strings are replaced by Unicode in 3.0"),
    ),
    Slide("Find Big Dict",
        Bullet("Util:", PYCODE("""
            def find_big_dict(x):
                import gc
                for obj in gc.get_referrers(x):
                    if type(obj) is dict and len(obj) > 1000:
                        return obj
                return None
        """)),
        Bullet("Example:", PYCON("""
            >>> interned = find_big_dict("a")
        """)),
        Bullet("Game:", PYCON("""
            >>> import util
            >>> util.keyword_quiz()
        """)),
    ),
    Slide("Ints",
        Bullet("For efficiency, ints are not malloc'ed singly but in PyIntBlocks."),
        Bullet("Ints are replaced by Longs in 3.0 -- malloc list is gone."),
        Bullet(PYCON("""
            Python 2.6a3+ (trunk:64269M, Jun 14 2008, 11:36:56) 
            >>> util.memory()
            5849088.0
            >>> z = list(range(1000000))
            >>> util.memory()
            22380544.0
            >>> del z
            >>> util.memory()
            17879040.0
        """)),
        Bullet(PYCON("""
            Python 3.0a5+ (py3k:64080, Jun 10 2008, 18:22:21)
            >>> util.memory()
            9302016.0
            >>> z = list(range(1000000))
            >>> util.memory()
            29011968.0
            >>> del z
            >>> util.memory()
            9302016.0
        """)),
    ),
    Slide("Int (speed)",
        Bullet(PYCON("""
            $ ./python -m timeit "list(x for x in range(2**30,2**30+1000))"
            10000 loops, best of 3: 176 usec per loop
            $ ../py3k/python -m timeit "list(x for x in range(2**30,2**30+1000))"
            1000 loops, best of 3: 202 usec per loop
            $ ./python -m timeit "list(x for x in range(2**37,2**37+1000))"
            1000 loops, best of 3: 311 usec per loop
            $ ../py3k/python -m timeit "list(x for x in range(2**37,2**37+1000))"
            1000 loops, best of 3: 431 usec per loop
            $ ./python -m timeit "list(x for x in range(2**63,2**63+1000))"
            1000 loops, best of 3: 320 usec per loop
            $ ../py3k/python -m timeit "list(x for x in range(2**63,2**63+1000))"
            1000 loops, best of 3: 439 usec per loop
        """)),
    ),
    Slide("Floats",
        Bullet("Like ints in 2.6, only in 3.0 as well."),
        Bullet("You won't get your memory back."),
        Bullet("Maybe we could remove this, run some tests and submit a patch (checking bug tracker first, of course)?"),
        Bullet("Perhaps just use numpy?"),
    ),
    Slide("C API Documentation",
        Bullet("PyObject API:",
            SubBullet(
                Bullet("List of methods unchanged between 2.5, 2.6 and 3.0"),
            ),
        ),
        Bullet("New documentation!",
            SubBullet(
                Bullet("Interactive search -- in static HTML"),
                Bullet("JSON index"),
                Bullet(URL("2.6 Docs","file://%s/Doc/build/html/index.html" % PYTHON_TRUNK_CHECKOUT)),
            ),
        ),
    ),
)

if __name__ == "__main__":
    pygments_css = "pygment.css"
    pygments_style = "default"

    l.renderHTML("./html", "pyobjects-%d.html", css=["main.css", pygments_css])

    from pygments.formatters import get_formatter_by_name
    formatter = get_formatter_by_name("html", style=pygments_style)
    css_file = file("./html/%s" % pygments_css, "w")
    try:
        css_file.write(formatter.get_style_defs(''))
    finally:
        css_file.close()
